/* "eqcs_matrix_rep.cc"
 * Copyright (C) 1999 Peter Belkner <pbelkner@berlin.snafu.de>
 *
 * This file is part of the eqcs library. This library is free
 * software; you can redistribute it and/or modify it under the terms of
 * the GNU General Public License as published by the Free Software
 * Foundation; either version 2, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; see the file COPYING.  If not, write to the
 * Free Software Foundation, 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
#ifdef __GNUG__
#pragma implementation
#endif

#include <stdarg.h>
#include <iostream.h>
#include "eqcs_sep.h"
#include "eqcs_state_rep.h"
#include "eqcs_matrix_rep.h"

/*****************************************************************************\
*	matrix representation													  *
\*****************************************************************************/

EqcsMatrixRep::EqcsMatrixRep(int nqubits, va_list ap) :
	_matrix(1 << nqubits), base_t(nqubits)
{
	for (int i = 0; i < _matrix.size(); ++i)
		EqcsBody::lock(_matrix[i] = new EqcsMatrixRowRep(this, nqubits));

	for (int i = 0; i < nqubits; ++i)
		// Get the next bit out of the variable parameter list.
		set(i, va_arg(ap, int));
}

EqcsMatrixRep::~EqcsMatrixRep()
{
	for (int i = 0; i < _matrix.size(); ++i)
		EqcsBody::unlock(_matrix[i]);
}

EqcsMatrixRowRep *EqcsMatrixRep::get(int row) const
{
	assert(0 <= row && row < _matrix.size());

	return _matrix[row];
}

EqcsStateRep *EqcsMatrixRep::apply(const EqcsStateRep *state)
{
	typedef EqcsMatrixRowRep::row_t row_t;
	typedef EqcsStateRep::term_t term_t;

	// Compute the result state's number of qubits.
	int n = maxbit() + 1;

	if (n < state->nqubits())
		n = state->nqubits();

	// Create an empty result state.
	EqcsStateRep *ret = new EqcsStateRep(n, 0);

	// Compute the result state's coefficients.
	// dim overflows!!!
	for (bitvec_t i = 0; i < ret->dim(); ++i) {
		// The coefficient c'[i] we're going to compute.
		complex_t c;

		// The matrix row corresponding to the current coefficient.
		bitvec_t row = mask2row(i);

		// Calculate the coefficient c'[i].
		for (bitvec_t j = 0; j < dim(); ++j) {
			// Get the matrix element m[row][j].
			if (0.0 == _matrix[row]->_row[j])
				continue;

			// Get the original coefficient c[k].
			bitvec_t col = col2mask(j);
			bitvec_t k = (i & ~bits()) | col;
			term_t::const_iterator p = state->_term.find(k);

			if (p == state->_term.end() || 0.0 == p->second)
				continue;

			// Add the term m[row][j]*c[k] to the coefficient c'[i].
			c += _matrix[row]->_row[j] * p->second;
		}

		if (0.0 != c)
			ret->_term[i] = c;
	}

	return ret;
}

ostream &EqcsMatrixRep::put(ostream &os) const
{
	typedef matrix_t::const_iterator const_iterator;

	EqcsSep sep("\n");

	for (const_iterator p = _matrix.begin(); p != _matrix.end(); ++p)
		(*p)->put(os << sep);

	return os << flush;
}

/*****************************************************************************\
*	matrix row representation												  *
\*****************************************************************************/

EqcsMatrixRowRep::EqcsMatrixRowRep(EqcsMatrixRep *matrix, int nqubits) :
	_row(1 << nqubits)
	{ _matrix = matrix; }

EqcsMatrixRowRep::~EqcsMatrixRowRep()
	{}

EqcsMatrixRowRep::complex_t &EqcsMatrixRowRep::get(int col)
{
	assert(0 <= col && col < _row.size());

	return _row[col];
}

ostream &EqcsMatrixRowRep::put(ostream &os) const
{
	EqcsSep sep("\t");

	for (row_t::const_iterator p = _row.begin(); p != _row.end(); ++p)
		os << sep << *p;

	return os << flush;
}
